﻿using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using TcpMessages;

namespace AdvCallManager
{ 
    class CallCenterServer
    {
        //public static CallDispatcher RepDispatcher = new CallDispatcher();
        private CallDispatcher _repDispatcher;
        private Socket _listener;
        SocketAsyncEventArgs _socketAcceptArgs = new SocketAsyncEventArgs();
        public CallCenterServer()
        {
            _repDispatcher = new CallDispatcher();
            _socketAcceptArgs.Completed += new EventHandler<SocketAsyncEventArgs>(AcceptAsyncCallback);

        }
        public void Start(int port)
        {
            IPHostEntry ipHostInfo = Dns.GetHostEntry(Dns.GetHostName());
            //Should be within the port range of 4502-4532
            IPEndPoint localEndPoint = new IPEndPoint(IPAddress.Any, port);

            _listener = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

            // Bind the socket to the local endpoint and listen for incoming connections.
            try
            {
                _listener.Bind(localEndPoint);
                _listener.Listen(50);
                AcceptAsync();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
        }
        void AcceptAsync()
        {
            AcceptAsync(_socketAcceptArgs);
        }

        void AcceptAsync(SocketAsyncEventArgs socketAcceptArgs)
        {

            if (!_listener.AcceptAsync(socketAcceptArgs))
            {
                AcceptAsyncCallback(socketAcceptArgs.AcceptSocket, socketAcceptArgs);
            }
        }

        void  AcceptAsyncCallback(object sender, SocketAsyncEventArgs e)
        {
            if (e.SocketError == SocketError.Success)
            {
                Console.WriteLine("Connection accepted...");
                RepConnection repCon = new RepConnection(e.AcceptSocket);
                repCon.OnReceive +=new EventHandler<ReceiveArgs>(OnReceiveCallback);
                //schedule the first receive
                repCon.ReceiveAsync();
            }
            //the following statement is necessary for the reuse of _socketAcceptArgs
            e.AcceptSocket = null;
            //schedule a new accept request
            AcceptAsync();
        }

        void OnReceiveCallback(object sender, ReceiveArgs e)
        {
            RepConnection repCon = (RepConnection)sender;
            ProcessMessage(e.Result, repCon);
        }

        public void ProcessMessage(NetworkMessage msg, RepConnection repCon)
        {
            switch(msg.Action)
            {
                case MessageAction.Register:
                    {
                        RegisterMessage registerMsg = (RegisterMessage)msg;
                        CallRep rep = new CallRep(repCon);
                        DateTime ts = DateTime.Now;
                        rep.RepName = registerMsg.RepName;
                        rep.RepNumber = registerMsg.RepNumber;
                        rep.PhoneExtension = registerMsg.PhoneExtension;
                        rep.ShiftStartTime = ts;
                        rep.Available = true;
                        //<todo> make thread safety changes
                        _repDispatcher.RepList.Add(rep.RepNumber, rep);
                        RegisterResponse registerResponse = new RegisterResponse(MessageAction.RegisterResponse);
                        registerResponse.RepNumber = registerMsg.RepNumber;
                        registerResponse.RegistrationTimestamp = ts;
                        registerResponse.CallManagerServerName = Dns.GetHostName();
                        Console.WriteLine("sending resonse...");
                        rep.RepConnection.SendAsync(registerResponse);
                    }
                    break;
                case MessageAction.Unregister:
                    {
                        RepKeyMessage unregisterMsg = (RepKeyMessage)msg;
                        string repKey = unregisterMsg.RepNumber;
                        CallRep rep = _repDispatcher.RepList[repKey];
                        if (rep != null)
                        {
                            lock (_repDispatcher.RepList)
                            {
                                _repDispatcher.RepList.Remove(repKey);
                                Console.WriteLine("Unregistering {0}, {1}, {2}", rep.RepName, rep.RepNumber, rep.PhoneExtension);
                            }
                        }
                    }
                    break;
                case MessageAction.Accept:
                    RepKeyMessage acceptMsg = msg as RepKeyMessage;
                    CallRep r1 = _repDispatcher.RepList[acceptMsg.RepNumber];
                    r1.Available = false;
                    r1.Locked = false;
                    r1.LockTime = DateTime.MaxValue;
                    r1.LastRejectTime = DateTime.MaxValue;
                    
                    break;
                case MessageAction.Reject:
                    RepKeyMessage rejectMsg = msg as RepKeyMessage;
                    CallRep r2 = _repDispatcher.RepList[rejectMsg.RepNumber];
                    //keep it locked for some time to avoid resending the request to the same rep
                    r2.Available = true;
                    r2.Locked = false;
                    r2.LockTime = DateTime.MaxValue;
                    r2.LastRejectTime = DateTime.Now;
                    break;
                //this will only true for the simulator
                case MessageAction.Call:
                    //receive the call
                    CallMessage callMsg = msg as CallMessage;
                    Console.WriteLine(string.Format("Received {0}'s call on {1}", callMsg.CallerName, callMsg.CallerPhone));
                    PhoneCall phoneCall = new PhoneCall();
                    phoneCall.Name = callMsg.CallerName;
                    phoneCall.PhoneNumber = callMsg.CallerPhone;
                    phoneCall.OrderNumber = callMsg.IvrData;
                    _repDispatcher.CustomerCallQueue.Enqueue(phoneCall);
                    break;
                case MessageAction.FinishCall:
                    RepKeyMessage finishCallMsg = msg as RepKeyMessage;
                    CallRep finishCallRep = _repDispatcher.RepList[finishCallMsg.RepNumber];
                    if (finishCallRep != null)
                    {
                        finishCallRep.Available = true;
                        finishCallRep.Locked = false;
                        finishCallRep.LockTime = DateTime.MaxValue;
                        finishCallRep.LastRejectTime = DateTime.MaxValue;
                        finishCallRep.NoOfCallsServiced++;
                        Console.WriteLine("Finishing the call {0}, {1}, {2}", finishCallRep.RepName, finishCallRep.RepNumber, finishCallRep.PhoneExtension);
                    }
                    break;
                default:
                    throw new ApplicationException(string.Format("Unknowns message type {0}",msg.Action.ToString()));
            }
        }
        public void Stop()
        {    
            //notify all the connected reps
            _repDispatcher.CleanUp();
            _listener.Close();
        }
    }
/// <summary>
/// RepConnection represents the accepted socket and the associated receive buffer. 
/// The CallDispatcher will maintain a list of RepConnection objects with live sockets 
/// for pushing notifications. RepConnection does not implement the client disconnect fucntionality.
/// </summary>

    class RepConnection
    {
        private ReceiveBuffer _receiveBuffer;
        private Socket _acceptedSocket; 
        public event EventHandler<ReceiveArgs> OnReceive;
        private SocketAsyncEventArgs _receiveSocketArgs;

        public ReceiveBuffer ReceiveBuffer
        {
            get { return _receiveBuffer; }
            set { _receiveBuffer = value; }
        }
        public RepConnection(Socket acceptedSocket)
        {
            this._acceptedSocket = acceptedSocket;
            this.ReceiveBuffer = new ReceiveBuffer();
            this._receiveSocketArgs = new SocketAsyncEventArgs();
            this._receiveSocketArgs.SetBuffer(_receiveBuffer.Buffer, 0, _receiveBuffer.Buffer.Length);
            _receiveSocketArgs.Completed += new EventHandler<SocketAsyncEventArgs>(ReceiveCallback); 
        }
        public Socket AcceptedSocket 
        {
            get { return _acceptedSocket; }
            set { _acceptedSocket = value; }
        }
        //receives a complete message
        public NetworkMessage ReceiveSync()
        {
            while (true)
            {
                int bytesReceived = _acceptedSocket.Receive(ReceiveBuffer.Buffer, ReceiveBuffer.Offset, ReceiveBuffer.Remaining, SocketFlags.None);
                //increment the offset
                _receiveBuffer.Offset += bytesReceived;

                //check if a complete message is already received
                if (_receiveBuffer.IsMessageReceived())
                {
                    break;
                }
            }
            //deseriaize the message
            NetworkMessage msg = NetworkMessage.Deserialize(ReceiveBuffer.Buffer);
            //adjust the buffer
            ReceiveBuffer.AdjustBuffer();
            return msg;
        }

        public void ReceiveAsync()
        {
            ReceiveAsync(_receiveSocketArgs);
        }

        private void ReceiveAsync(SocketAsyncEventArgs socketAsyncEventArgs)
        {
            if (!_acceptedSocket.ReceiveAsync(socketAsyncEventArgs))
            {
                ReceiveCallback(_acceptedSocket, socketAsyncEventArgs);
            }
        }

        void ReceiveCallback(object sender, SocketAsyncEventArgs e)
        {
            if (e.SocketError != SocketError.Success)
            {
                return;
            }
            _receiveBuffer.Offset += e.BytesTransferred;
            if (_receiveBuffer.IsMessageReceived())
            {
                if (OnReceive != null)
                {
                    NetworkMessage msg = NetworkMessage.Deserialize(_receiveBuffer.Buffer);
                    _receiveBuffer.AdjustBuffer();
                    OnReceive(this, new ReceiveArgs(msg));
                }
            }
            else
            {
                //adjust the buffer pointer
                e.SetBuffer(_receiveBuffer.Offset, _receiveBuffer.Remaining);
            }
            //queue a an async read request
            ReceiveAsync(_receiveSocketArgs);
        }

        public void SendSync(NetworkMessage msg)
        {
            int numBytes = _acceptedSocket.Send(NetworkMessage.Serialize(msg));
        }

        public void SendAsync(NetworkMessage msg)
        {
            SocketAsyncEventArgs sendSocketArgs = new SocketAsyncEventArgs();
            byte[] sendBytes = NetworkMessage.Serialize(msg);
            sendSocketArgs.SetBuffer(sendBytes, 0, sendBytes.Length);
            sendSocketArgs.Completed += new EventHandler<SocketAsyncEventArgs>(SendCallback);
            SendAsync(sendSocketArgs);
        }

        private void SendAsync(SocketAsyncEventArgs sendSocketArgs)
        {
            if (!_acceptedSocket.SendAsync(sendSocketArgs))
            {
                //call is completed synchronously
                SendCallback(_acceptedSocket, sendSocketArgs);
            }
        }

        void SendCallback(object sender, SocketAsyncEventArgs e)
        {
            if (e.BytesTransferred < e.Count)
            {
                //some more bytes to send
                e.SetBuffer(e.BytesTransferred, e.Count - e.BytesTransferred);
                SendAsync(e);
            }
        }
    }
    public class ReceiveArgs : EventArgs
    {
        private NetworkMessage _msg = null;
        public ReceiveArgs(NetworkMessage msg)
        {
            this._msg = msg;
        }
        public NetworkMessage Result
        {
            get { return _msg; }
        }
    }

}
