﻿using System;
using System.Net;
using System.Net.Sockets;
using System.Windows;
using TcpMessages;

namespace AdvClientLibrary
{
    public class ReceiveArgs : EventArgs
    {
        private NetworkMessage _msg = null;
        public ReceiveArgs(NetworkMessage msg)
        {
            this._msg = msg;
        }
        public NetworkMessage Result
        {
            get { return _msg; }
        }
    }
    public class CallNetworkClient
    {
        private Socket _socket;
        private ReceiveBuffer _receiveBuffer;

        public event EventHandler<EventArgs> OnConnectError;
        public event EventHandler<ReceiveArgs> OnReceive;
        public SocketAsyncEventArgs _receiveSocketArgs;
        public SocketAsyncEventArgs _sendSocketArgs;

        public CallNetworkClient()
        {
            _receiveBuffer = new ReceiveBuffer();
            _receiveSocketArgs = new SocketAsyncEventArgs();
            _sendSocketArgs = new SocketAsyncEventArgs();
            _receiveSocketArgs.SetBuffer(_receiveBuffer.Buffer, _receiveBuffer.Offset, _receiveBuffer.Remaining);
            _receiveSocketArgs.Completed += new EventHandler<SocketAsyncEventArgs>(ReceiveCallback);
            _sendSocketArgs.Completed += new EventHandler<SocketAsyncEventArgs>(SendCallback);
        }
        public SocketAsyncEventArgs RecieveSocketArgs
        {
            get { return _receiveSocketArgs; }
        }
        public SocketAsyncEventArgs SendSocketArgs
        {
            get { return _sendSocketArgs; }
        }
        public Socket ConnectedSocket
        {
            get { return _socket; }
        }
        public void Connect(int portNo)
        {
            _socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            SocketAsyncEventArgs connectEventArgs = new SocketAsyncEventArgs();
            connectEventArgs.Completed += new EventHandler<SocketAsyncEventArgs>(connectEventArgs_Completed);
            connectEventArgs.RemoteEndPoint = new DnsEndPoint(Application.Current.Host.Source.Host, portNo);
            _socket.ConnectAsync(connectEventArgs);
        }
        void connectEventArgs_Completed(object sender, SocketAsyncEventArgs e)
        {
            if (e.SocketError == SocketError.Success)
            {
                //drop a receive request in the thread pool 
                ReceiveAsync();
            }
            else
            {
                if (OnConnectError != null)
                    OnConnectError(sender, e);
            }
        }

        public void ReceiveAsync()
        {
           ReceiveAsync(_receiveSocketArgs);
        }

        private void ReceiveAsync(SocketAsyncEventArgs recvArgs)
        {
            if (!_socket.ReceiveAsync(recvArgs))
            {
                ReceiveCallback(_socket, recvArgs);
            }
        }

        void ReceiveCallback(object sender, SocketAsyncEventArgs e)
        {
            if (e.SocketError != SocketError.Success)
            {
                return;
            }

            _receiveBuffer.Offset += e.BytesTransferred;
            if (_receiveBuffer.IsMessageReceived())
            {
                if (OnReceive != null)
                {
                    NetworkMessage msg = NetworkMessage.Deserialize(_receiveBuffer.Buffer);
                    _receiveBuffer.AdjustBuffer();
                    OnReceive(this, new ReceiveArgs(msg));
                }
            }
            else
            {
                //adjust the buffer pointer
                e.SetBuffer(_receiveBuffer.Offset, _receiveBuffer.Remaining);
            }
            //drop a receive request into the threadpool
            ReceiveAsync();
        }

        public void SendAsync(NetworkMessage msg)
        {
            byte[] sendBytes = NetworkMessage.Serialize(msg);
            _sendSocketArgs.SetBuffer(sendBytes, 0, sendBytes.Length);
            SendAsync(_sendSocketArgs);
        }

        private void SendAsync(SocketAsyncEventArgs sendSocketArgs)
        {
            if (!_socket.SendAsync(sendSocketArgs))
            {
                //call is completed synchronously
                SendCallback(this._socket, sendSocketArgs);
            }
        }

        void SendCallback(object sender, SocketAsyncEventArgs e)
        {
            if (e.SocketError != SocketError.Success)
            {
                //do error management here
                return;
            }
            if (e.BytesTransferred < e.Count)
            {
                //send is not complete more bytes to send
                e.SetBuffer(e.BytesTransferred, e.Count - e.BytesTransferred);
                SendAsync(e);
            }
        }

    }
}