﻿//CAUTION:This group of classes will be used on the Silverlight Client also
//so, please make sure of the compatibility at the source code level
using System;
using System.Text;
using System.Xml.Serialization;
using System.IO;

namespace TcpMessages
{
    public enum MessageAction
    {
        InvalidAction,
        Register,
        RegisterResponse,
        Unregister,
        UnRegisterResponse,
        Call,
        FinishCall,
        Accept,
        Reject
    }
    [XmlInclude(typeof(RegisterMessage))]
    [XmlInclude(typeof(RegisterResponse))]
    [XmlInclude(typeof(CallMessage))]
    [XmlInclude(typeof(RepKeyMessage))]
    public abstract class NetworkMessage
    {
        public static readonly int LENGTH_BYTES = 4;
        protected MessageAction _messageAction;
        public MessageAction Action 
        { 
            get { return _messageAction; }
            set { _messageAction = value; }
        }
        //first 4 bytes of the buffer will contain the length of the serialized NetworkMessage
        //this function treats input argument as immutable
        public static NetworkMessage Deserialize(byte[] buffer)
        {
            int length = BitConverter.ToInt32(buffer, 0);
            int startIndex = sizeof(int);
            string strNetworkMessage = Encoding.UTF8.GetString(buffer, startIndex, length);

            StringReader stringReader = new StringReader(strNetworkMessage);
            XmlSerializer xmlSerializer = new XmlSerializer(typeof(NetworkMessage));
            return (NetworkMessage)xmlSerializer.Deserialize(stringReader);
        }

        //the first 4 bytes of the serialized byte array will contain the length of the buffer that
        //contains the previously serialized NetworkMessage;
        public static byte[] Serialize(NetworkMessage msg)
        {
            StringWriter stringWriter = new StringWriter();
            XmlSerializer xmlSerializer = new XmlSerializer(typeof(NetworkMessage));
            xmlSerializer.Serialize(stringWriter, msg);
            byte[] messageBytes = Encoding.UTF8.GetBytes(stringWriter.ToString());
            byte[] lengthBytes = BitConverter.GetBytes(messageBytes.Length);
            byte[] finalBytes = new byte[lengthBytes.Length + messageBytes.Length];
            //copy the length of the serialized object
            Array.Copy(lengthBytes, 0, finalBytes, 0, lengthBytes.Length);
            Array.Copy(messageBytes, 0, finalBytes, lengthBytes.Length, BitConverter.ToInt32(lengthBytes, 0));
            return finalBytes;
        }
 
    }

    public class RegisterMessage : NetworkMessage
    {
        public RegisterMessage()
        {
            _messageAction = MessageAction.InvalidAction;
        }
        public RegisterMessage(MessageAction action)
        {
            _messageAction = action; 
        }

        public string RepName { get; set; }
        public string RepNumber { get; set; }
        public int PhoneExtension { get; set; }

    }
    public class RegisterResponse : NetworkMessage
    {
         public RegisterResponse()
        {
            _messageAction = MessageAction.InvalidAction;
        }
         public RegisterResponse(MessageAction action)
        {
            _messageAction = action; 
        }
        public string RepNumber { get; set; }
        public string CallManagerServerName { get; set; }
        public DateTime RegistrationTimestamp { get; set; }
    }

    public class CallMessage : NetworkMessage
    {
        public CallMessage()
        {
            _messageAction = MessageAction.InvalidAction;
        }
        public CallMessage(MessageAction action)
        {
            _messageAction = action; 
        }

        public string CallerName { get; set; }
        public string CallerPhone { get; set; }
        public string IvrData { get; set; }
    }

    public class RepKeyMessage : NetworkMessage
    {
        public RepKeyMessage()
        {
            _messageAction = MessageAction.InvalidAction;
        }
        public RepKeyMessage(MessageAction action)
        {
            _messageAction = action;
        }
        public string RepNumber { get; set; }
    }

    class SocketBuffer
    {
        public const int BUFFERSIZE = 32768;
        protected byte[] _buffer = new byte[BUFFERSIZE];
        protected int _offset = 0;

        public byte[] Buffer
        {
            get { return _buffer; }
            set { _buffer = value; }
        }
        //offset will allways indicate the length of the buffer that is filled
        public int Offset
        {
            get {return _offset ;}
            set { _offset = value; }
        }

        public int Remaining
        {
            get { return _buffer.Length - _offset; }
        }
    }

    class ReceiveBuffer : SocketBuffer
    {
        //removes a serlialized message from the buffer, copies the partial message to the begining
        //and adjusts the offset
        public void AdjustBuffer()
        {
            int messageSize = BitConverter.ToInt32(_buffer, 0);
            int lengthToCopy = _offset - NetworkMessage.LENGTH_BYTES - messageSize;
            Array.Copy(_buffer, _offset, _buffer, 0, lengthToCopy);
            _offset = lengthToCopy;
        }
        //this method checks if a complete message is received
        public bool IsMessageReceived()
        {
            if (_offset < 4)
                return false;
            int sizeToRecieve = BitConverter.ToInt32(_buffer, 0);
            //check if we have a complete NetworkMessage
            if ((_offset - 4) < sizeToRecieve)
                return false; //we have not received the complete message yet
            //we received the complete message and may be more
            return true;
        }
    }
}
